/*******************
* Danmaku engine
*******************/
#ifndef THBULLETS_H
#define THBULLETS_H

#include <PA9.h>
#define FUNCPTR void*

//#include "thgoodness.h"

#define BULLETS_MAX 500		//maximum number of bullets 
#define BULLETS_FIRSTSPRITE 5	//first sprite to use for bullets

#define BULLET_FIRSTROT 4	//first available rotation set
#define BULLET_MAXROTSETS 25	//

//Bullet flags
#define BULLET_ACTIVE	0x01	//if this bullet is in use or not
//#define BULLET_ROUND 	0x02	//if this item is to be considered round (else its considered rectangle)
// ALL TOUHOU HITBOXEN ARE ROUND! thus the safe vs hit area brillant!
#define BULLET_HIT	0x02	//this bullet has been already detected as a hit/graze so dont check again
#define BULLET_FOLLOW	0x04	//If this flag is set bullet will not update and instead just follow the master sprite with lcnt
				//also xinc and yinc can be used for an offset ??
#define BULLET_SPIN	0x08	//if this sprite should spin about an axis
#define BULLET_STICKY	0x10	//this bullet should not be reused or terminate ?
#define BULLET_WIDE	0x20	//if clear the bullet is destroyed immediately after it exits the
				//screen, if set will wait till right before its about to wrap around :)
#define BULLET_GRAZE	0x40	//this bullet has already been detected as a graze
#define BULLET_PHANTOM	0x80	//this bullet has no sprite
#define BULLET_ISPLAYER 0x100	//this bullet is the player itself!


//Passed to bulletCreate
struct badv_s {
	int lcnt;	//how long (in frames) do i live?
	int mdly;	//how long (in frames) before i start moving
	int msdly;	//how long (in frames) before i stop moving?
	FUNCPTR func;	//what custom function do i use?
	int follow;	//who do i follow?  SHOULD BE PARENT
	u8  priority;	//what priority is this sprite
	u8  alpha;	//enable alpha blending?
};

// Always passed to bulletCreate (simplifies creating multiple bullets with same properties)
struct bsprite_s {
	u8    shapex;
	u8    shapey;
	u8    pal;
	u8    size;
	signed char    hitrad;
};


//Radial bullet effects (to do like spinning arms of stuff)
struct brad_s {
	int	radius;
	int	angle;
	int	radspd;
	int	anglespd;
};


//bullets structure (one of these for each projectile sprite) or mebbe even just each sprite
struct t_bullets {
	s32 x;	//starting xpos
	s32 y;	//starting ypos
	int realx;
	int realy;	//actual finally drawn bullet positions (functions or radials may change from center x,y)
	int radius;	//radial radius (from x,y) <--- SHOULD BE ORBITAL!
	int angle;	//rotation radial angle 
	int radspd;	//speed to increase the radius
	int anglespd;	//speed to rotate the angle
	u8  rotset;	//which rotation set to associate with this
	int xinc;	//how much to increment each time for x and y, divided by 256 for real pos
	int yinc;	
	int size;	//size (shoule be square) for calculating center of sprite
	int hrad;	//size of hitbox for this item (-1 = no hitbox)
	u32 flags;	//various boolian flags?
	s32 xvfl;	//x and y virtual fixed point values (basically x <<8 or x*256)
	s32 yvfl;	
	int sprite;	//sprite used for this object	
	int who;		//who does this object belong to? (-1 = player, anything else = enemy and points to an enemy struct)
	int follow;	//follows this bullet SHOULD BE PARENTs
	int lcnt;	//bullet only lives for this many frames then destroys itself
	int mdly;	//delay this many frames b4 allowing movement
	int msdly;	//delay till i stop moving
	int t1;
	int ax1;	//acceleration effects!!
	int ay1;
	int t2;
	int ax2;	//similar  but the time here is when it starts 
	int ay2;	//whew!!! hope this will be it this is a rather massive struct 43KBytes @ 400 bullets
	FUNCPTR func;	//function for 'smart bullets'
};

typedef int(*_pBulletFunc)(int,int);	//Bullet function pointer type

//This is referenced other places so export it
extern struct t_bullets bullets[];

//Function prototypes
void bulletInit(void);
int bulletCreate(void *sprite, struct bsprite_s *bsprite, int x, int y, int xinc, int yinc, int rotset, struct badv_s *badv, struct brad_s *brad, u32 flags, int who);
//int bulletCreate(void *sprite, struct bsprite_s *bsprite, int x, int y, int xinc, int yinc, int rotset, struct badv_s *badv);
void bulletUpdate(void);
void bulletsDestroyAllNonSticky(void);
void bulletsDestroyAll(void);
void bulletsReplace(int bullet, void *obj, int size, int newwho,u8 pal);
void bulletsReplaceAllSizeDrop(void *obj, int size, int newwho,u8 who);
void bulletsReplaceAllWhoSizeDrop(void *obj, int who, int size, int newwho,u8 pal);
void bulletsDestroyAllSize(int size);
void bulletsDestroyAllWhoSize(int who, int size);
void bulletsDestroyAllWho(int who);
void bulletsDestroyAllWhoSticky(int who);
void bulletsDestroyAllFollow(int who); //might be better to just always do this..
void bulletsDestroy(int bullet);
void bulletsSetAllWhoAlpha(int who, u8 alpha);
u32 bulletsSetFlags(int bullet,u32 flags);
u32 bulletsClearFlags(int bullet,u32 flags);
void bulletsSetAcceleration(int bullet, int t1, int a1, int t2, int a2);
void bulletsSetAccelerationCoefs(int bullet, int t1, int ax1, int xy1, int t2, int ax2, int ay2);




//bulletsSetXinc(bullet, ((grimoire[curspell].moves[spstates[who].mvndx].x<<8)-(bulletsGetRealX(bullet)<<8))/grimoire[curspell].moves[spstates[who].mvndx].frms);

//  does (xb<<8)-(xa<<8) == (ab-xa)<<8 ??
#define bulletsSetVector(b,x,y,f) bullets[(b)].xinc=((((x)<<8)-(bullets[(b)].realx<<8))/(f)); bullets[(b)].yinc=((((y)<<8)-(bullets[(b)].realy<<8))/(f));  
//Here thar be useful macros! (looks prettier than using the raw bullet scuct stuffs outside the bullet lib and is FAST
#define bulletsSetAlpha(b,a) (thSpriteSetAlpha(bullets[(b)].sprite,(a)))
#define bulletsGetX(b) (bullets[b].x)
#define bulletsGetY(b) (bullets[b].y)	//note setting .x or .y is pointless as its only used for reference outside bulletlib
					//i could just not use it in bullet lib and make macros to do the math when needed
					//instead of bullet lib updating values used by outside functions
#define bulletsSetMSdly(b,d) (bullets[(b)].msdly=(d))
#define bulletsSetXY(b,x,y) (bullets[(b)].xvfl=(x)<<8,bullets[(b)].yvfl=(y)<<8)
#define bulletsSetX(b,x,y) (bullets[b].xvfl=(x)<<8)
#define bulletsSetXvfl(b,x) (bullets[b].xvfl=(x))
#define bulletsSetYvfl(b,y) (bullets[b].yvfl=(y))
#define bulletsSetXinc(b,xin) (bullets[b].xinc=(xin))
#define bulletsSetYinc(b,yin) (bullets[b].yinc=(yin))
#define bulletsSetRadius(b,s) (bullets[b].radius=(s<<8))
#define bulletsGetXinc(b) (bullets[b].xinc)
#define bulletsGetYinc(b) (bullets[b].yinc)
#define bulletsGetRealX(b) (bullets[b].realx)
#define bulletsGetRealY(b) (bullets[b].realy)
#define bulletsGetSize(b) (bullets[b].size)
#define bulletsGetHitrad(b) (bullets[b].hrad)
#define bulletsGetWho(b) (bullets[b].who)
#define bulletsGetFlags(b) (bullets[b].flags)

#endif
