/*
	thdualsprites.c - purpose here is to create dual screen sprite engine (cause palib's sux :/)
		should have the following features:
			~ dynamically created/destroyed sprites
			~ efficient use of hardware sprites only allocating sprite for which screen it can be viewed on
			~ auto deallocation of hardware sprites that are not visible and reallocation if returned to screen
			~ loose priority system, tries to give u whatever priority u request or closest match.. -1 = psuedo random priority
*/

#include <PA9.h>
#include "thdualsprites.h"
#include "thgoodness.h"
#include "thbullets.h"




//global for nextthsprite

int thnextsprite=0;

struct thSprite_s thSprites[THMAXVSPRITES];	//table of virtual sprites :D


//These will be used as flags to indicate its being used... and 2 can indicate sticky(wont be cleared offscreen)
u8 spritetab[2][THMAXSPRITES];	//bottom screen (pardon the soft drink pun :P)


//returns 0 if on bottm, 1 if on top and -1 if not visable on either, 2 if on both!?
inline int thSpriteChk(int x, int y, u8 size) {
	if(x>256 || y>384 || (x<-size) || (y<-size)) {	// sprite outta bounds
		return -1;
	}
	if((y<=191) && (y+size>191)) return 2; //both, gah i almost forgot this (well till i saw it only on one or the other hehe)
	if(y>191) return 0; 	//bottom screen
	return 1;		//top screen
}


//This is gonna be slow use a last sprite type deal perhaps? would be much faster...
inline int thGetNextSprite(u8 screen) {
	int c=0;
	do {
		if(spritetab[screen][c]==0) {
			spritetab[screen][c]=1;
			return c;
		}
	} while(++c!=THMAXSPRITES);
	return -1; //cant alloc?! u done run outta sprites,y0!
}


//Initalize thSprite system
void thSpriteInit(void) {
	 int c=0;
	do {
		thSprites[c].used=0;
		thSprites[c].hwsprite0=-1;
		thSprites[c].hwsprite1=-1;
		thSprites[c].priority=-1;
	} while(++c!=THMAXVSPRITES);
	c=0;
	do {
		spritetab[0][c]=0;
		spritetab[1][c]=0;
	} while(++c!=THMAXSPRITES);
	thnextsprite=0;
}

//creates a sprite, or two.. or not depends... (used internally)
inline int thSubCreateSprite(struct thSprite_s *thsp,int screen,int x, int y) {
	int hwsp;
//	PA_OutputText(1, 4, 6, "screen %d  ",screen);
	if(((screen==2) || (screen==0 )) && (thsp->hwsprite0<0)) {
		if(thsp->priority>=0) {
			hwsp=thsp->priority;
		} else {
			hwsp=thGetNextSprite(0);
		}
		if(hwsp>=0) { //can u gimme a sprite on this screen?
			if(y>(191-thsp->size)) y-=192;
//			PA_OutputText(1, 4, 4, "sc %d sp %d",thsp,hwsp);
			PA_CreateSprite(0,hwsp,thsp->sprite, thsp->shapex,thsp->shapey,1,thsp->pal,x,y);
			PA_SetSpritePrio(0, hwsp, 1);
//it may be faster to check ->alpha first.... will see during efficiency pass :)
			PA_SetSpriteMode(0, hwsp, thsp->alpha); // Alphablending

			if(thsp->rotset>=0) {
				PA_SetSpriteRotEnable(0,hwsp, // screen,sprite number
				      thsp->rotset);// rotset number. You have 32 rotsets (0-31) per screen. 2 sprites with*
			}
		}
		thsp->hwsprite0=hwsp;
	}
	if(((screen==2) || (screen==1 )) && (thsp->hwsprite1<0)) {
		if(thsp->priority>=0) {
			hwsp=thsp->priority;
		} else {
			hwsp=thGetNextSprite(1);
		}
		if(hwsp>=0) { //can u gimme a sprite on this screen?
//			PA_OutputText(1, 4, 1, "sc %d sp %d",thsp,hwsp);
			PA_CreateSprite(1,hwsp,thsp->sprite, thsp->shapex,thsp->shapey,1,thsp->pal,x,y);
			PA_SetSpritePrio(1, hwsp, 1);
			PA_SetSpriteMode(1, hwsp, thsp->alpha); // Alphablending
			if(thsp->rotset>=0) {
				PA_SetSpriteRotEnable(1,hwsp, // screen,sprite number
				      thsp->rotset);// rotset number. You have 32 rotsets (0-31) per screen. 2 sprites with*
			}
		}
		thsp->hwsprite1=hwsp;
	}
	return 0;
}

//Finds the hw sprite most closely matching the requested priority
signed char thSpriteGetNearestPrio(signed char priority) {
	int c=0;
	do {
		if((c+priority<THMAXSPRITES) && (spritetab[0][priority+c]==0) && (spritetab[1][priority+c]==0)) return priority+c;
		if((priority-c>=0) && (spritetab[0][priority-c]==0) && (spritetab[1][priority-c]==0)) return priority-c;
	} while(++c<127);
	return -1;	//okai i giveup give it random prio!
}

//int testcnt=0;

/*
* create a sprite
	returns sprite handle or -1 if error...
*/

// does priority need to be set here? can i have an external macro to set it??
int thSpriteCreate(void *sprite, u8 shapex, u8 shapey, int size, u8 pal,int x, int y, int rotset, signed char priority) {
	int c=0;
//	int thsp;
	while(thSprites[thnextsprite].used!=0) {
		if(++thnextsprite==THMAXVSPRITES)
			thnextsprite=0;
	}
	c=thnextsprite;
	thSprites[c].alpha=0;	//disable alpha blending by default
	thSprites[c].used=1;
	thSprites[c].sprite=sprite;
	thSprites[c].shapex=shapex;
	thSprites[c].shapey=shapey;
	thSprites[c].size=size;
	thSprites[c].pal=pal;
	thSprites[c].rotset=rotset;
	if(priority>=THMAXSPRITES) priority=THMAXSPRITES-1;
	if(priority>=0) {
		if((spritetab[0][priority]==1) || (spritetab[1][priority]==1)) {
				//Find closest match
//			PA_OutputText(0, 4, 1, "requested    : %d",priority);
			priority=thSpriteGetNearestPrio(priority);
//			PA_OutputText(0, 4, 2, "nearest match: %d",priority);
		//		priority=-1;
		}
		spritetab[0][priority]=1;	//mark both in use always
		spritetab[1][priority]=1;
//		PA_OutputText(0, 2, 3, "priority request cnt: %d",++testcnt);

	//priority=-1;

	}
	thSprites[c].priority=priority;
	thSprites[c].screen=-1;
	thSpriteSetXY(c, x, y);
	return(c);	//return sprite handle
}


//Check if these sprites need to be deleted cause of screen changes
inline void thRelSprite(struct thSprite_s *thsp, int screen) {
	if((thsp->hwsprite0>=0) && ((screen<0) || (screen==1))) {
		PA_DeleteSprite(0, thsp->hwsprite0);
//	PA_OutputText(0, 4, 4, "destroyed a 0 %d  ",thsp->hwsprite0);
		if(thsp->priority<0) {
			spritetab[0][thsp->hwsprite0]=0;
		}
		thsp->hwsprite0=-1; 

	}
	if((thsp->hwsprite1>=0) && ((screen<0) || (screen==0))) {
		PA_DeleteSprite(1, thsp->hwsprite1);
		if(thsp->priority<0) {
			spritetab[1][thsp->hwsprite1]=0;
		}
		thsp->hwsprite1=-1; 
//	PA_OutputText(0, 4, 4, "destroyed a 1 ");
	}

}

//Sets new x,y position first arg= sprite handle
int thSpriteSetXY(int spriteh, int x, int y) {
	int thsp;
	thsp=thSpriteChk(x,y,thSprites[spriteh].size);
	if(thsp != thSprites[spriteh].screen) {
//		if(!(thSprites[spriteh].priority)) {	//if sprite has priority set then it should try to create or release here
			thSubCreateSprite(&thSprites[spriteh],thsp,x, y);
			thRelSprite(&thSprites[spriteh],thsp);
//		}
		thSprites[spriteh].screen=thsp;	//new screen now save it
	}
	if(thsp==1 || thsp==2) PA_SetSpriteXY(1, thSprites[spriteh].hwsprite1, x, y);
	if(y>(191-thSprites[spriteh].size)) y-=192;
	if(thsp==0 || thsp==2) PA_SetSpriteXY(0, thSprites[spriteh].hwsprite0, x, y);		
	return spriteh;
}


//prints debug info about how many sprites are in use on which screens etc etc...
void thSpriteDiag(void) {
	int c=0;
	int s0s=0;
	int s1s=0;
	do {
		s0s+=spritetab[0][c];
		s1s+=spritetab[1][c];
	}while(++c!=THMAXSPRITES);
	PA_OutputText(0, 4, 20, "used0: %d   used1: %d    ",s0s,s1s);
	c=0;
	s0s=0;
	do {
		if(thSprites[c].used>0) s0s++;
	}while(++c!=THMAXVSPRITES);
	PA_OutputText(0, 4, 21, "vused: %d   next: %d    ",s0s,thnextsprite);
}

//called by external functions to destory a created sprite, pass spriteHandle
int thSpriteDestroy(int spriteh) {
	if((thSprites[spriteh].priority) || (thSprites[spriteh].screen>=0)) { // if we had prio set need to destory 
		thSprites[spriteh].priority=-1;
		thRelSprite(&thSprites[spriteh],-1);
	}
	thSprites[spriteh].used=0;
	thSprites[spriteh].hwsprite0=-1;
	thSprites[spriteh].hwsprite1=-1;
	return(0);
}

//right so we enable alpha here...
void thSpriteSetAlpha(int spriteh, unsigned char alpha){
		thSprites[spriteh].alpha=alpha;
		if(thSprites[spriteh].hwsprite0>=0) PA_SetSpriteMode(0,thSprites[spriteh].hwsprite0, alpha); // Alphablending
		if(thSprites[spriteh].hwsprite1>=0) PA_SetSpriteMode(1,thSprites[spriteh].hwsprite1, alpha); // Alphablending
}

//replace sprite's pallet
void thSpriteUpdatePal(int spriteh, u8 pal) {
	if(thSprites[spriteh].hwsprite0>=0)
	 	PA_SetSpritePal(0,thSprites[spriteh].hwsprite0,FAMPAL); 	 
	if(thSprites[spriteh].hwsprite1>=0) 
		PA_SetSpritePal(1,thSprites[spriteh].hwsprite1,FAMPAL);
	thSprites[spriteh].pal=FAMPAL;
}

//update the image used by the sprite (must be same pallete, same type, same size etc...)
void thSpriteUpdateGfx(int spriteh,void *obj) {
	if(thSprites[spriteh].sprite!=obj) {	//This being here helps a bit :D
		thSprites[spriteh].sprite=obj;
		if(thSprites[spriteh].hwsprite0>=0) 
			PA_UpdateSpriteGfx(0,thSprites[spriteh].hwsprite0 , obj);
		if(thSprites[spriteh].hwsprite1>=0) 
			PA_UpdateSpriteGfx(1,thSprites[spriteh].hwsprite1, obj);
	}
}

//update the image used by the sprite (must be same pallete, same type, same size etc...)
void thSpriteUpdateGfxPal(int spriteh,void *obj,u8 pal) {
	if(thSprites[spriteh].sprite!=obj) {	//This being here helps a bit :D
		thSprites[spriteh].sprite=obj;
		if(thSprites[spriteh].hwsprite0>=0) {
			PA_UpdateSpriteGfx(0,thSprites[spriteh].hwsprite0 , obj);
			PA_SetSpritePal(0,thSprites[spriteh].hwsprite0,pal); 	 
		}
		if(thSprites[spriteh].hwsprite1>=0) {
			PA_UpdateSpriteGfx(1,thSprites[spriteh].hwsprite1, obj);
			PA_SetSpritePal(1,thSprites[spriteh].hwsprite1,pal);

 		}
	thSprites[spriteh].pal=pal;
	}
}


void thSpriteUpdateRotset(int spriteh,int rotset) {
	thSprites[spriteh].rotset=rotset;
	if(rotset>=0) {
		if(thSprites[spriteh].hwsprite0>=0)
			PA_SetSpriteRotEnable(0, // screen,sprite number
			      thSprites[spriteh].hwsprite0,rotset);// rotset number. You have 32 rotsets (*
		if(thSprites[spriteh].hwsprite1>=0)
			PA_SetSpriteRotEnable(1, // screen,sprite number
			      thSprites[spriteh].hwsprite1,rotset);// rotset number. You have 32 rotsets (*
	} else {
		if(thSprites[spriteh].hwsprite0>=0)
			PA_SetSpriteRotDisable(0, thSprites[spriteh].hwsprite0);
		if(thSprites[spriteh].hwsprite1>=0)
			PA_SetSpriteRotDisable(1, thSprites[spriteh].hwsprite1);
	}
}



/*
 These are the rotlib functions!
*/
struct brot_s throts[ROTSETMAX];

void rotsetInit(void) {
	int c=0;
	do {
		throts[c].used=0;
	} while(++c!=ROTSETMAX);
}

void rotsetUpdate() {
	int c=0;
	do {
		if(throts[c].used) {
			throts[c].angle+=throts[c].spd;
			PA_DualSetRotsetNoZoom(c, // rotset
			   (throts[c].angle>>8)&511); // angle, from 0 to 511
		}
	} while(++c!=ROTSETMAX);
}

u8 rotsetCreate(int spd, u32 angle) {
	int c=0;
	while(throts[c].used) {
		if(++c==ROTSETMAX) return -1;
	};
	throts[c].used=1;
	throts[c].spd=spd;
	throts[c].angle=angle;
	return(c);
}

void rotsetDestroy(u8 rotset) {
	throts[rotset].used=0;
}
